.view, viewport, list {
 background: transparent;
}

paneldockchild.start {
 background: color-mix(in srgb, var(--border-color), transparent);
}

paneldockchild.start revealer.top-bar {
 /* background: transparent; */
 border-bottom: 1px solid var(--border-color);
}

panelresizer > widget > separator.vertical {
 background-color: var(--sidebar-bg-color);
 background-image: image(var(--sidebar-border-color));
}

paneldockchild.start entry {
 min-height: 29px;
 font-size: 0.83333em;
 box-shadow: inset 0 1px var(--sidebar-border-color);
 border-radius: 0;
}

paneldockchild.start entry:dir(ltr) {
 border-bottom-left-radius: var(--window-radius);
}

paneldockchild.start entry:dir(rtl) {
 border-bottom-right-radius: var(--window-radius);
}

panelstatusbar {
 min-height: 28px;
}
