paneltogglebutton.drag-active button:not(:checked) {
  background: color-mix(in srgb, var(--accent-bg-color) 25%, transparent);
}

panelframeheaderbar {
  min-height: 38px;
  padding-top: 1px;
  padding-bottom: 3px;
  padding-left: 3px;
}
panelframe panelframeheaderbar,
paneldockchild.center panelgrid > panelpaned > panelresizer:first-child:last-child > panelgridcolumn > panelpaned panelresizer:first-child:last-child panelframeheaderbar {
  border-top: 2px solid transparent;
  border-bottom: 1px solid color-mix(in srgb, var(--border-color) 60%, transparent);
}
panelframe.has-focus panelframeheaderbar{
  border-top: 2px solid var(--accent-bg-color);
}
panelframeheaderbar > box {
  margin-left: 2px;
  margin-right: 2px;
}
panelframeheaderbar button:not(:hover):not(:checked) {
  background: transparent;
}
panelframeheaderbar button {
  margin: 1px;
  font-weight: normal;
  border-radius: 9px;
}
panelgrid panelframeheaderbar .controls {
  margin-right: 4px;
}
panelgrid panelresizer:last-child > panelgridcolumn panelframeheaderbar .controls {
  /* align with window close button on right-most */
  margin-right: 8px;
}
panelframeheaderbar .controls button.circular {
  padding: 0px;
  margin: 0px;
  min-height: 24px;
  min-width: 24px;
  border-radius: 9999px;
}
panelframeheaderbar .controls button.circular image {
  margin-left: 4px;
  margin-right: 4px;
}

paneldockchild.end panelpaned.vertical panelresizer:first-child panelframe panelframeswitcher,
paneldockchild.start panelpaned.vertical panelresizer:first-child panelframe panelframeswitcher {
  border-bottom: 1px solid color-mix(in srgb, var(--border-color) 60%, transparent);
}
panelpaned panelframe panelframeswitcher:backdrop {
  filter: opacity(.5);
  transition: 200ms ease-in-out;
}
panelpaned.horizontal panelframe panelframeswitcher {
  padding: 2px;
  min-width: 40px;
}
panelpaned.vertical panelframe panelframeswitcher {
  padding: 2px;
  min-height: 40px;
}
panelpaned.vertical panelframe panelframeswitcher button {
  padding: 2px 6px;
  background-clip: padding-box;
}
panelpaned.horizontal panelframe panelframeswitcher button {
  padding: 6px 2px;
  background-clip: padding-box;
}
panelframe panelframeswitcher button {
  border-radius: 13px;
  /* use border instead of margin so that there are no gaps
   * when dragging across buttons. This helps things feel a
   * bit more solid when doing DnD.
   */
  border: 3px solid transparent;
  background-clip: padding-box;
}
panelframe panelframeswitcher button:not(:hover):not(:checked):not(:active) {
  background: transparent;
  background-clip: padding-box;
  color: color-mix(in srgb, currentColor 75%, transparent);
}
panelframeswitcher.drop-indicator {
  color: color-mix(in srgb, var(--accent-bg-color) 80%, transparent);
}

panelframe .frameheader .action-bar popover.menu.pages-popover row {
  padding: 0 6px 0 10px;
  border-radius: 6px;
  min-width: 40px;
  min-height: 32px;
}
panelframe .frameheader .action-bar popover.menu.pages-popover row:selected:not(:hover) {
  background: transparent;
}
panelframe .frameheader .action-bar popover.menu.pages-popover row check {
  opacity: 0;
}
panelframe .frameheader .action-bar popover.menu.pages-popover row:selected check {
  opacity: 1;
}
panelframe .frameheader .action-bar popover.menu.pages-popover row .close-button {
  padding: 0;
  margin: 3px 0;
  min-width: 24px;
  min-height: 24px;
  border-radius: 99px;
  border: none;
  box-shadow: none;
  background: none;
  color: transparent;
}
panelframe .frameheader .action-bar popover.menu.pages-popover row:hover .close-button {
  color: var(--window-fg-color);
}
panelframe .frameheader .action-bar popover.menu.pages-popover row .close-button:hover {
  background: rgb(0 0 0 / 8%);
}
panelframe .frameheader .action-bar popover.menu.pages-popover row .close-button:active {
  background: rgb(0 0 0 / 15%);
}
panelframe .frameheader tab > button.tab-close-button {
  min-width: 24px;
  min-height: 24px;
}


panelstatusbar {
  min-height: 24px;
  font-size: 0.9em;
  padding-top: 1px;
  box-shadow: inset 0 1px var(--headerbar-shade-color);
  background: var(--headerbar-bg-color);
  color: var(--headerbar-fg-color);
}
panelstatusbar:backdrop {
  background: var(--headerbar-backdrop-color);
  transition: 200ms ease-in-out;
}
panelstatusbar > * {
  border-right: 1px solid var(--border-color);
  padding-left: 9px;
  padding-right: 9px;
}
panelstatusbar:backdrop > * {
  filter: opacity(.5);
  transition: 200ms ease-in-out;
}
panelstatusbar > *:last-child {
  border-right: none;
}
panelstatusbar paneltogglebutton,
panelstatusbar > menubutton,
panelstatusbar > dropdown {
  padding: 0;
}
panelstatusbar paneltogglebutton button,
panelstatusbar > dropdown > button,
panelstatusbar > menubutton > button,
panelstatusbar > button {
  min-width: 8px;
  border-radius: 0;
  padding: 0 9px;
}

panelstatusbar {
  --start-left-radius: var(--window-radius);
  --start-right-radius: var(--window-radius);
  --end-left-radius: var(--window-radius);
  --end-right-radius: var(--window-radius);
}

panelstatusbar:dir(ltr) {
  --start-right-radius: 0;
  --end-left-radius: 0;
}

panelstatusbar:dir(rtl) {
  --start-left-radius: 0;
  --end-right-radius: 0;
}

paneldock.start-revealed paneldockchild.bottom panelstatusbar:dir(ltr),
paneldock.end-revealed paneldockchild.bottom panelstatusbar:dir(ltr) {
  --start-left-radius: 0;
  --end-right-radius: 0;
}

paneldock.start-revealed paneldockchild.bottom panelstatusbar:dir(rtl),
paneldock.end-revealed paneldockchild.bottom panelstatusbar:dir(rtl) {
  --start-right-radius: 0;
  --end-left-radius: 0;
}

panelstatusbar > button:first-child,
panelstatusbar > dropdown:first-child > button,
panelstatusbar > menubutton:first-child > button,
panelstatusbar > paneltogglebutton:first-child button {
  border-bottom-left-radius: var(--start-left-radius);
  border-bottom-right-radius: var(--start-right-radius);
}

panelstatusbar > button:last-child,
panelstatusbar > dropdown:last-child > button,
panelstatusbar > menubutton:last-child > button,
panelstatusbar > paneltogglebutton:last-child button {
  border-bottom-left-radius: var(--end-left-radius);
  border-bottom-right-radius: var(--end-right-radius);
}

panelstatusbar paneltogglebutton button,
panelstatusbar > dropdown > button,
panelstatusbar > menubutton > button,
panelstatusbar > button {
  background: transparent;
}
panelstatusbar paneltogglebutton button:hover,
panelstatusbar > dropdown > button:hover,
panelstatusbar > menubutton > button:hover,
panelstatusbar > button:hover {
  background: color-mix(in srgb, currentColor 7%, transparent);
}
panelstatusbar paneltogglebutton button:active,
panelstatusbar paneltogglebutton button.keyboard-activating,
panelstatusbar > dropdown > button:hover,
panelstatusbar > dropdown > button.keyboard-activating,
panelstatusbar > menubutton > button:hover,
panelstatusbar > menubutton > button.keyboard-activating,
panelstatusbar > button:hover,
panelstatusbar > button.keyboard-activating {
  background: color-mix(in srgb, currentColor 16%, transparent);
}
panelstatusbar paneltogglebutton button:checked,
panelstatusbar > dropdown > button:checked,
panelstatusbar > menubutton > button:checked,
panelstatusbar > button:checked {
  background: color-mix(in srgb, currentColor 10%, transparent);
}
panelstatusbar paneltogglebutton button:checked:hover,
panelstatusbar > dropdown > button:checked:hover,
panelstatusbar > menubutton > button:checked:hover,
panelstatusbar > button:hover {
  background: color-mix(in srgb, currentColor 13%, transparent);
}
panelstatusbar paneltogglebutton button:checked:active,
panelstatusbar paneltogglebutton button:checked.keyboard-activating,
panelstatusbar > dropdown > button:checked:hover,
panelstatusbar > dropdown > button:checked.keyboard-activating,
panelstatusbar > menubutton > button:checked:hover,
panelstatusbar > menubutton > button:checked.keyboard-activating,
panelstatusbar > button:hover,
panelstatusbar > button.keyboard-activating {
  background: color-mix(in srgb, currentColor 19%, transparent);
}
panelstatusbar > dropdown > button label,
panelstatusbar > menubutton > button label,
panelstatusbar > button label {
  font-weight: normal;
}

/* Style statusbar correctly within a toolbar view */
.top-bar panelstatusbar,
.bottom-bar panelstatusbar {
  padding-top: 0px;
  box-shadow: none;
  background: none;
  color: inherit;
}
.top-bar panelstatusbar:backdrop,
.bottom-bar panelstatusbar:backdrop {
  background: none;
  transition: none;
}
.top-bar panelstatusbar:backdrop > *,
.bottom-bar panelstatusbar:backdrop > * {
  filter: none;
  transition: none;
}

panelframe panelframetabbar {
  border-bottom: 1px solid color-mix(in srgb, var(--border-color) 60%, transparent);
  min-height: 44px;
}
panelframe panelframetabbar:backdrop {
  color: color-mix(in srgb, currentColor, transparent);
}
panelframe panelframetabbar > overlay > box.focus-handle,
paneldockchild.center panelgrid > panelpaned > panelresizer:first-child:last-child > panelgridcolumn > panelpaned panelresizer:first-child:last-child panelframetabbar box.focus-handle {
  min-height: 2px;
  background: transparent;
}
panelframe.has-focus panelframetabbar > overlay > box.focus-handle {
  background: var(--accent-bg-color);
}
panelframe panelframetabbar tabbox tab button {
  min-height: 0;
}
panelframe panelframetabbar button.frame-close-button {
  min-width: 28px;
  min-height: 28px;
}
panelframe panelframetabbar tabbox {
  padding: 0;
}
panelframe panelframetabbar tabbox tabboxchild {
  margin-top: 5px;
  margin-bottom: 5px;
}
panelframe panelframetabbar .end-action {
  padding: 5px;
}

.omnibar > splitbutton > button {
  margin-bottom: 0;
  margin-top: 0;
  padding-bottom: 0;
  padding-top: 0;
}


/* frame switcher buttons */
panelframeheaderbar button {
  background-color: transparent;
}
panelframeheaderbar button:checked,
panelframeswitcher button:checked {
 background-color: color-mix(in srgb, currentColor 12.5%, transparent);
}
panelframeheaderbar button:hover,
panelframeswitcher button:hover {
 background-color: color-mix(in srgb, currentColor 10%, transparent);
}
panelframeheaderbar button:hover:checked,
panelframeswitcher button:hover:checked {
 background-color: color-mix(in srgb, currentColor 15%, transparent);
}
panelframeheaderbar button:active:hover,
panelframeswitcher button:active:hover {
 background-color: color-mix(in srgb, currentColor 25%, transparent);
}



panelthemeselector {
  margin: 9px;
}
panelthemeselector checkbutton {
  padding: 0;
  min-height: 44px;
  min-width: 44px;
  padding: 1px;
  background-clip: content-box;
  border-radius: 9999px;
  box-shadow: inset 0 0 0 1px var(--border-color);

  --light-bg: #fff;
  --dark-bg: #202020;
}
panelthemeselector checkbutton.follow:checked,
panelthemeselector checkbutton.light:checked,
panelthemeselector checkbutton.dark:checked {
  box-shadow: inset 0 0 0 2px var(--accent-bg-color);
}
panelthemeselector checkbutton.follow {
  background-image: linear-gradient(to bottom right, var(--light-bg) 49.99%, var(--dark-bg) 50.01%);
}
panelthemeselector checkbutton.light {
  background-color: var(--light-bg);
}
panelthemeselector checkbutton.dark {
  background-color: var(--dark-bg);
}
panelthemeselector checkbutton radio {
  -gtk-icon-source: none;
  border: none;
  background: none;
  box-shadow: none;
  min-width: 12px;
  min-height: 12px;
  transform: translate(27px, 14px);
  padding: 2px;
}
panelthemeselector checkbutton radio:checked {
  -gtk-icon-source: -gtk-icontheme("object-select-symbolic");
  background-color: var(--accent-bg-color);
  color: var(--accent-fg-color);
}


.omnibar > splitbutton {
  background: color-mix(in srgb, currentColor 7.5%, transparent);
}


panelframe:drop(active) {
  box-shadow: none;
}
panelframe paneldropcontrols {
  transition-property: background, border;
  transition-duration: 300ms;
  border-radius: 10px;
  margin: 1px;
}
panelframe paneldropcontrols {
  background: color-mix(in srgb, var(--accent-bg-color) 10%, transparent);
}
panelframe paneldropcontrols:drop(active) {
  background: color-mix(in srgb, var(--accent-bg-color) 30%, transparent);
  box-shadow: inset 0 0 0 1px var(--accent-bg-color);
}
paneldropcontrols button {
  border-radius: 11px;
  color: color-mix(in srgb, var(--accent-fg-color) 80%, transparent);
  background-color: var(--accent-bg-color);
  box-shadow: 0 3px 5px rgba(0 0 0 / 20%);
  padding: 0;
}
paneldropcontrols button.circular {
  border-radius: 9999px;
}
paneldropcontrols button:drop(active) {
  background-color: color-mix(in srgb, var(--accent-bg-color), var(--accent-fg-color) 30%);
  color: var(--accent-fg-color);
}

panelprogressicon {
  color: var(--accent-color);
}

panelframeswitcher button {
  background-clip: padding-box;
}
